@@ -1,3 +1,5 @@
+open Misc
+
 module Path : sig
   (** [to_shortest_lid ~env ~env_check path] will make a [Longident.t] from the
   provided [Path.t] and attempt to use the shortest prefix possible given the
@@ -22,3 +24,23 @@ end
 
 (* Add parenthesis to qualified operators *)
 val parenthesize_name : string -> string
+
+(** [parse_identifier] attempts to re-parse a longident so that we get
+    the location of each of its components. *)
+val parse_identifier :
+  (Mconfig.t * Msource.t) -> Lexing.position -> modname Location.loc list
+
+module Compat : sig
+  val pat_var_id_and_loc :
+    Typedtree.pattern -> (Ident.t * string Location.loc) option
+
+  val pat_alias_pat_id_and_loc
+    : Typedtree.pattern
+    -> (Typedtree.pattern * Ident.t * string Location.loc) option
+end
+
+(** Extracts the location of a [uid] from a [Typedtree.item_declaration] *)
+val loc_of_decl :
+  uid:Shape.Uid.t ->
+  Typedtree.item_declaration ->
+  string Location.loc option
