@@ -6,6 +6,7 @@ let {Logger. log} = Logger.for_section "Mconfig"
 
 type ocaml = {
   include_dirs         : string list;
+  hidden_dirs          : string list;
   no_std_include       : bool;
   unsafe               : bool;
   classic              : bool;
@@ -15,7 +16,6 @@ type ocaml = {
   recursive_types      : bool;
   strict_sequence      : bool;
   applicative_functors : bool;
-  unsafe_string        : bool;
   nopervasives         : bool;
   strict_formats       : bool;
   open_modules         : string list;
@@ -32,6 +32,7 @@ let dump_warnings st =
 
 let dump_ocaml x = `Assoc [
     "include_dirs"         , `List (List.map ~f:Json.string x.include_dirs);
+    "hidden_dirs"          , `List (List.map ~f:Json.string x.hidden_dirs);
     "no_std_include"       , `Bool x.no_std_include;
     "unsafe"               , `Bool x.unsafe;
     "classic"              , `Bool x.classic;
@@ -40,7 +41,6 @@ let dump_ocaml x = `Assoc [
     "recursive_types"      , `Bool x.recursive_types;
     "strict_sequence"      , `Bool x.strict_sequence;
     "applicative_functors" , `Bool x.applicative_functors;
-    "unsafe_string"        , `Bool x.unsafe_string;
     "nopervasives"         , `Bool x.nopervasives;
     "strict_formats"       , `Bool x.strict_formats;
     "open_modules"         , Json.list Json.string x.open_modules;
@@ -72,11 +72,16 @@ let marg_commandline f =
 type merlin = {
   build_path  : string list;
   source_path : string list;
+  hidden_build_path  : string list;
+  hidden_source_path : string list;
   cmi_path    : string list;
   cmt_path    : string list;
+  index_files : string list;
   extensions  : string list;
   suffixes    : (string * string) list;
   stdlib      : string option;
+  source_root : string option;
+  unit_name   : string option;
   reader      : string list;
   protocol    : [`Json | `Sexp];
   log_file    : string option;
@@ -104,8 +109,11 @@ let dump_merlin x =
   `Assoc [
     "build_path"   , `List (List.map ~f:Json.string x.build_path);
     "source_path"  , `List (List.map ~f:Json.string x.source_path);
+    "hidden_build_path" , `List (List.map ~f:Json.string x.hidden_build_path);
+    "hidden_source_path", `List (List.map ~f:Json.string x.hidden_source_path);
     "cmi_path"     , `List (List.map ~f:Json.string x.cmi_path);
     "cmt_path"     , `List (List.map ~f:Json.string x.cmt_path);
+    "index_files"  , `List (List.map ~f:Json.string x.index_files);
     "flags_applied", `List (List.map ~f:dump_flag_list x.flags_applied);
     "extensions"   , `List (List.map ~f:Json.string x.extensions);
     "suffixes"     , `List (
@@ -115,6 +123,8 @@ let dump_merlin x =
         ]) x.suffixes
     );
     "stdlib"       , Json.option Json.string x.stdlib;
+    "source_root"  , Json.option Json.string x.source_root;
+    "unit_name"    , Json.option Json.string x.unit_name;
     "reader"       , `List (List.map ~f:Json.string x.reader);
     "protocol"     , (match x.protocol with
         | `Json -> `String "json"
@@ -232,6 +242,33 @@ let rec normalize t =
   ) else
     normalize (normalize_step t)
 
+let merge_merlin_config dot merlin ~failures ~config_path =
+  { merlin with
+    build_path = dot.Mconfig_dot.build_path @ merlin.build_path;
+    source_path = dot.source_path @ merlin.source_path;
+    hidden_build_path = dot.hidden_build_path @ merlin.hidden_build_path;
+    hidden_source_path = dot.hidden_source_path @ merlin.hidden_source_path;
+    cmi_path = dot.cmi_path @ merlin.cmi_path;
+    cmt_path = dot.cmt_path @ merlin.cmt_path;
+    index_files = dot.index_files @ merlin.index_files;
+    exclude_query_dir = dot.exclude_query_dir || merlin.exclude_query_dir;
+    use_ppx_cache = dot.use_ppx_cache || merlin.use_ppx_cache;
+    extensions = dot.extensions @ merlin.extensions;
+    suffixes = dot.suffixes @ merlin.suffixes;
+    stdlib = (if dot.stdlib = None then merlin.stdlib else dot.stdlib);
+    source_root =
+      (if dot.source_root = None then merlin.source_root else dot.source_root);
+    unit_name =
+      (if dot.unit_name = None then merlin.unit_name else dot.unit_name);
+    reader =
+      if dot.reader = []
+      then merlin.reader
+      else dot.reader;
+    flags_to_apply = dot.flags @ merlin.flags_to_apply;
+    failures = failures @ merlin.failures;
+    config_path = Some config_path;
+  }
+
 let get_external_config path t =
   let path = Misc.canonicalize_filename path in
   let directory = Filename.dirname path in
@@ -239,26 +276,7 @@ let get_external_config path t =
   | None -> t
   | Some (ctxt, config_path) ->
     let dot, failures = Mconfig_dot.get_config ctxt path in
-    let merlin = t.merlin in
-    let merlin = {
-      merlin with
-      build_path = dot.build_path @ merlin.build_path;
-      source_path = dot.source_path @ merlin.source_path;
-      cmi_path = dot.cmi_path @ merlin.cmi_path;
-      cmt_path = dot.cmt_path @ merlin.cmt_path;
-      exclude_query_dir = dot.exclude_query_dir || merlin.exclude_query_dir;
-      use_ppx_cache = dot.use_ppx_cache || merlin.use_ppx_cache;
-      extensions = dot.extensions @ merlin.extensions;
-      suffixes = dot.suffixes @ merlin.suffixes;
-      stdlib = (if dot.stdlib = None then merlin.stdlib else dot.stdlib);
-      reader =
-        if dot.reader = []
-        then merlin.reader
-        else dot.reader;
-      flags_to_apply = dot.flags @ merlin.flags_to_apply;
-      failures = failures @ merlin.failures;
-      config_path = Some config_path;
-    } in
+    let merlin = merge_merlin_config dot t.merlin ~failures ~config_path in
     normalize { t with merlin }
 
 let merlin_flags = [
@@ -274,6 +292,18 @@ let merlin_flags = [
         {merlin with source_path = dir :: merlin.source_path}),
     "<dir> Add <dir> to merlin source path"
   );
+  (
+    "-hidden-build-path",
+    marg_path (fun dir merlin ->
+        {merlin with hidden_build_path = dir :: merlin.hidden_build_path}),
+    "<dir> Add <dir> to merlin hidden build path"
+  );
+  (
+    "-hidden-source-path",
+    marg_path (fun dir merlin ->
+        {merlin with hidden_source_path = dir :: merlin.hidden_source_path}),
+    "<dir> Add <dir> to merlin hidden source path"
+  );
   (
     "-cmi-path",
     marg_path (fun dir merlin ->
@@ -286,6 +316,12 @@ let merlin_flags = [
         {merlin with cmt_path = dir :: merlin.cmt_path}),
     "<dir> Add <dir> to merlin cmt path"
   );
+  (
+    "-index-file",
+    marg_path (fun file merlin ->
+        {merlin with index_files = file :: merlin.index_files}),
+    "<file> Add <file> to the index files used by merlin"
+  );
   (
     "-reader",
     Marg.param "command" (fun reader merlin ->
@@ -429,7 +465,7 @@ let ocaml_ignored_parametrized_flags = [
   "-inline"; "-inline-prim-cost"; "-inline-toplevel"; "-intf";
   "-intf_suffix"; "-intf-suffix"; "-o"; "-rounds"; "-runtime-variant";
   "-unbox-closures-factor"; "-use-prims"; "-use_runtime"; "-use-runtime";
-  "-error-style"; "-dump-dir";
+  "-error-style"; "-dump-dir"; "-cmi-file";
 ]
 
 let ocaml_warnings_spec ~error =
@@ -457,6 +493,13 @@ let ocaml_flags = [
         {ocaml with include_dirs = dir :: ocaml.include_dirs}),
     "<dir> Add <dir> to the list of include directories"
   );
+  (
+    "-H",
+    marg_path (fun dir ocaml ->
+        {ocaml with hidden_dirs = dir :: ocaml.hidden_dirs}),
+    "<dir>  Add <dir> to the list of \"hidden\" include directories\n\
+    \ (Like -I, but the program can not directly reference these dependencies)"
+  );
   (
     "-nostdlib",
     Marg.unit (fun ocaml -> {ocaml with no_std_include = true}),
@@ -517,19 +560,10 @@ let ocaml_flags = [
     Marg.unit (fun ocaml -> {ocaml with threads = `None}),
     " Add support for VM-scheduled threads library"
   );
-  (
-    "-unsafe-string",
-    Marg.unit (fun ocaml -> {ocaml with unsafe_string = true}),
-    Printf.sprintf
-      " Make strings mutable (default: %B)"
-      (not Config.safe_string)
-  );
   (
     "-safe-string",
-    Marg.unit (fun ocaml -> {ocaml with unsafe_string = false}),
-    Printf.sprintf
-      " Make strings immutable (default: %B)"
-      Config.safe_string
+    Marg.unit (fun ocaml -> ocaml),
+    " Default to true unconditionally since 5.00"
   );
   (
     "-nopervasives",
@@ -598,6 +632,7 @@ let ocaml_flags = [
 let initial = {
   ocaml = {
     include_dirs         = [];
+    hidden_dirs          = [];
     no_std_include       = false;
     unsafe               = false;
     classic              = false;
@@ -607,7 +642,6 @@ let initial = {
     recursive_types      = false;
     strict_sequence      = false;
     applicative_functors = true;
-    unsafe_string        = not Config.safe_string;
     nopervasives         = false;
     strict_formats       = false;
     open_modules         = [];
@@ -618,11 +652,16 @@ let initial = {
   merlin = {
     build_path  = [];
     source_path = [];
+    hidden_build_path  = [];
+    hidden_source_path = [];
     cmi_path    = [];
     cmt_path    = [];
+    index_files = [];
     extensions  = [];
     suffixes    = [(".ml", ".mli"); (".re", ".rei")];
     stdlib      = None;
+    source_root = None;
+    unit_name   = None;
     reader      = [];
     protocol    = `Json;
     log_file    = None;
@@ -731,7 +770,8 @@ let source_path config =
   List.concat
     [[config.query.directory];
      stdlib;
-     config.merlin.source_path]
+     config.merlin.source_path;
+     config.merlin.hidden_source_path]
   |> List.filter_dup
 
 let build_path config = (
@@ -763,6 +803,9 @@ let build_path config = (
   result'
 )
 
+let hidden_build_path config =
+  config.merlin.hidden_build_path @ config.ocaml.hidden_dirs
+
 let cmt_path config = (
   let dirs =
     match config.ocaml.threads with
@@ -773,6 +816,7 @@ let cmt_path config = (
   let dirs =
     config.merlin.cmt_path @
     config.merlin.build_path @
+    config.merlin.hidden_build_path @
     dirs
   in
   let stdlib = stdlib config in
@@ -795,4 +839,7 @@ let global_modules ?(include_current=false) config = (
 
 let filename t = t.query.filename
 
-let unitname t = Misc.unitname t.query.filename
+let unitname t =
+  match t.merlin.unit_name with
+  | Some name -> Misc.unitname name
+  | None -> Misc.unitname t.query.filename
