@@ -27,9 +27,3 @@
 )* }}} *)
 
 val fresh_env : unit -> Env.t
-
-val rewrite_loc :
-  [ `str of Parsetree.structure | `sg of Parsetree.signature
-  | `fake of Parsetree.structure ] ->
-  [ `str of Parsetree.structure | `sg of Parsetree.signature
-  | `fake of Parsetree.structure ]
