@@ -6,15 +6,17 @@
 
 (** {1 Relevant settings}
     Parameters from OCaml compiler which affect Merlin behavior. *)
+val cmi_file             : string option ref
 val include_dirs         : string list ref
+val hidden_include_dirs  : string list ref
 val fast                 : bool ref
 val classic              : bool ref
+val all_ppx              : string list ref
 val principal            : bool ref
 val real_paths           : bool ref
 val recursive_types      : bool ref
 val strict_sequence      : bool ref
 val applicative_functors : bool ref
-val unsafe_string        : bool ref
 val nopervasives         : bool ref
 val strict_formats       : bool ref
 val open_modules         : string list ref
@@ -23,6 +25,7 @@ val open_modules         : string list ref
     Ignored by merlin but kept for compatibility with upstream code. *)
 val annotations          : bool ref
 val binary_annotations   : bool ref
+val store_occurrences    : bool ref
 val print_types          : bool ref
 val native_code          : bool ref
 val dont_write_files     : bool ref
